@@ -2,7 +2,7 @@ package Mail::Milter::Authentication::Handler::IPRev;
 use strict;
 use warnings;
 use base 'Mail::Milter::Authentication::Handler';
-use version; our $VERSION = version->declare('v1.1.5');
+use version; our $VERSION = version->declare('v1.1.6');
 
 use Net::DNS;
 use Net::IP;
@@ -65,12 +65,19 @@ sub connect_callback {
     my $ptr_list = {};
     my @error_list;
 
+    my @cname_hosts;
+
     my $packet = $resolver->query( $ip_address, 'PTR' );
     $lookup_limit--;
     if ($packet) {
         foreach my $rr ( $packet->answer ) {
-            next unless $rr->type eq "PTR";
-            $ptr_list->{ $rr->rdatastr } = [];
+            if ( $rr->type eq "CNAME" ) {
+                push @cname_hosts, $rr->rdatastr;
+                push @error_list, 'Found CNAME in PTR response';
+            }
+            if ( $rr->type eq "PTR" ) {
+                $ptr_list->{ $rr->rdatastr } = [];
+            }
         }
     }
 
@@ -79,24 +86,61 @@ sub connect_callback {
         push @error_list, 'Error ' . $resolver->errorstring() . " looking up $ip_address PTR";
     }
 
+    foreach my $cname_host ( @cname_hosts ) {
+        my $packet = $resolver->query( $cname_host, 'PTR' );
+        $lookup_limit--;
+        if ($packet) {
+            foreach my $rr ( $packet->answer ) {
+                #if ( $rr->type eq "CNAME" ) {
+                # NO! We only follow the first level CNAMES
+                # Because anything more is probably busted anyway
+                #}
+                if ( $rr->type eq "PTR" ) {
+                    $ptr_list->{ $rr->rdatastr } = [];
+                }
+            }
+        }
+        if ( $resolver->errorstring() ) {
+            $self->_dns_error( 'PTR', $cname_host, $resolver->errorstring );
+            push @error_list, 'Error ' . $resolver->errorstring() . " looking up $cname_host PTR";
+        }
+        last; # Because multiple CNAMES for a given record is also busted
+    }
+
     if ( ! keys %$ptr_list ) {
         push @error_list, "NOT FOUND";
     }
 
+    my @lookup_list = sort keys %$ptr_list;
     DOMAINLOOKUP:
-    foreach my $domain ( sort keys %$ptr_list ) {
-
+    foreach my $domain ( @lookup_list ) {
+    warn "***** looking up $domain\n\n";
         my $ip_list = [];
+        my $cname;
 
         if ( $ip_address =~ /:/ ) {
             # We are living in the future!
 
             my $errors6;
             my $errors4;
-            ( $lookup_limit, $ip_list, $errors6 ) = $self->_address_for_domain( 'AAAA', $domain, $lookup_limit );
+            ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'AAAA', $domain, $lookup_limit );
+            if ( $cname ) {
+                push @error_list, 'Found CNAME in AAAA response';
+                ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'AAAA', $cname, $lookup_limit );
+                if ( ! @$ip_list ) {
+                    ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'A', $cname, $lookup_limit );
+                }
+            }
             if ( ! @$ip_list ) {
                 # We got nothing, try ip4
-                ( $lookup_limit, $ip_list, $errors4 ) = $self->_address_for_domain( 'A', $domain, $lookup_limit );
+                ( $lookup_limit, $ip_list, $errors4, $cname ) = $self->_address_for_domain( 'A', $domain, $lookup_limit );
+                if ( $cname ) {
+                    push @error_list, 'Found CNAME in A response';
+                    ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'AAAA', $cname, $lookup_limit );
+                    if ( ! @$ip_list ) {
+                        ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'A', $cname, $lookup_limit );
+                    }
+                }
             }
             if ( ! @$ip_list ) {
                 foreach my $error ( @$errors4 ) {
@@ -112,10 +156,24 @@ sub connect_callback {
 
             my $errors6;
             my $errors4;
-            ( $lookup_limit, $ip_list, $errors4 ) = $self->_address_for_domain( 'A', $domain, $lookup_limit );
+            ( $lookup_limit, $ip_list, $errors4, $cname ) = $self->_address_for_domain( 'A', $domain, $lookup_limit );
+            if ( $cname ) {
+                push @error_list, 'Found CNAME in A response';
+                ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'A', $cname, $lookup_limit );
+                if ( ! @$ip_list ) {
+                    ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'AAAA', $cname, $lookup_limit );
+                }
+            }
             if ( ! @$ip_list ) {
                 # We got nothing, try ip6
-                ( $lookup_limit, $ip_list, $errors6 ) = $self->_address_for_domain( 'AAAA', $domain, $lookup_limit );
+                ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'AAAA', $domain, $lookup_limit );
+                if ( $cname ) {
+                    push @error_list, 'Found CNAME in AAAA response';
+                    ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'A', $cname, $lookup_limit );
+                    if ( ! @$ip_list ) {
+                        ( $lookup_limit, $ip_list, $errors6, $cname ) = $self->_address_for_domain( 'AAAA', $cname, $lookup_limit );
+                    }
+                }
             }
             if ( ! @$ip_list ) {
                 foreach my $error ( @$errors4 ) {
@@ -176,6 +234,7 @@ sub _address_for_domain {
 
     my @fwd_errors;
     my @ip_list;
+    my $cname;
 
     my $resolver = $self->get_object('resolver');
 
@@ -187,9 +246,14 @@ sub _address_for_domain {
 
     if ($packet) {
         foreach my $rr ( $packet->answer ) {
-            next unless $rr->type eq $type;
-            my $address = $rr->rdatastr;
-            push @ip_list, $address;
+            if (  lc $rr->type eq 'cname' ) {
+                $cname = $rr->rdatastr;
+                # Multiple CNAMES are broken, but we don't check for that
+                # We just take the last one we found
+            }
+            if ( lc $rr->type eq lc $type ) {
+                push @ip_list, $rr->rdatastr;
+            }
         }
     }
 
@@ -198,7 +262,7 @@ sub _address_for_domain {
         push @fwd_errors, 'Error ' . $resolver->errorstring() . " looking up $domain $type";
     }
 
-    return ( $lookup_limit, \@ip_list, \@fwd_errors );
+    return ( $lookup_limit, \@ip_list, \@fwd_errors, $cname );
 }
 
 sub close_callback {
